@@ -153,14 +153,14 @@ abstract class Kernel extends QMatrix {
 
 	private static double powi(double base, int times)
 	{
-	        double tmp = base, ret = 1.0;
+		double tmp = base, ret = 1.0;
 
-        	for(int t=times; t>0; t/=2)
+		for(int t=times; t>0; t/=2)
 		{
-                	if(t%2==1) ret*=tmp;
-	                tmp = tmp * tmp;
-        	}
-	        return ret;
+			if(t%2==1) ret*=tmp;
+			tmp = tmp * tmp;
+		}
+		return ret;
 	}
 
 	double kernel_function(int i, int j)
@@ -274,7 +274,7 @@ abstract class Kernel extends QMatrix {
 			case svm_parameter.SIGMOID:
 				return Math.tanh(param.gamma*dot(x,y)+param.coef0);
 			case svm_parameter.PRECOMPUTED:
-				return 	x[(int)(y[0].value)].value;
+				return	x[(int)(y[0].value)].value;
 			default:
 				return 0;	// java
 		}
@@ -316,7 +316,7 @@ class Solver {
 	int[] active_set;
 	double[] G_bar;		// gradient, if we treat free variables as 0
 	int l;
-	boolean unshrinked;	// XXX
+	boolean unshrink;	// XXX
 	
 	static final double INF = java.lang.Double.POSITIVE_INFINITY;
 
@@ -364,18 +364,40 @@ class Solver {
 
 		if(active_size == l) return;
 
-		int i;
-		for(i=active_size;i<l;i++)
-			G[i] = G_bar[i] + p[i];
+		int i,j;
+		int nr_free = 0;
 
-		for(i=0;i<active_size;i++)
-			if(is_free(i))
+		for(j=active_size;j<l;j++)
+			G[j] = G_bar[j] + p[j];
+
+		for(j=0;j<active_size;j++)
+			if(is_free(j))
+				nr_free++;
+
+		if(2*nr_free < active_size)
+			svm.info("\nWarning: using -h 0 may be faster\n");
+
+		if (nr_free*l > 2*active_size*(l-active_size))
+		{
+			for(i=active_size;i<l;i++)
 			{
-				float[] Q_i = Q.get_Q(i,l);
-				double alpha_i = alpha[i];
-				for(int j=active_size;j<l;j++)
-					G[j] += alpha_i * Q_i[j];
-			}
+				float[] Q_i = Q.get_Q(i,active_size);
+				for(j=0;j<active_size;j++)
+					if(is_free(j))
+						G[i] += alpha[j] * Q_i[j];
+			}	
+		}
+		else
+		{
+			for(i=0;i<active_size;i++)
+				if(is_free(i))
+				{
+					float[] Q_i = Q.get_Q(i,l);
+					double alpha_i = alpha[i];
+					for(j=active_size;j<l;j++)
+						G[j] += alpha_i * Q_i[j];
+				}
+		}
 	}
 
 	void Solve(int l, QMatrix Q, double[] p_, byte[] y_,
@@ -390,7 +412,7 @@ class Solver {
 		this.Cp = Cp;
 		this.Cn = Cn;
 		this.eps = eps;
-		this.unshrinked = false;
+		this.unshrink = false;
 
 		// initialize alpha_status
 		{
@@ -445,7 +467,7 @@ class Solver {
 			{
 				counter = Math.min(l,1000);
 				if(shrinking!=0) do_shrinking();
-				System.err.print(".");
+				svm.info(".");
 			}
 
 			if(select_working_set(working_set)!=0)
@@ -454,7 +476,7 @@ class Solver {
 				reconstruct_gradient();
 				// reset active set size and check
 				active_size = l;
-				System.err.print("*");
+				svm.info("*");
 				if(select_working_set(working_set)!=0)
 					break;
 				else
@@ -630,7 +652,7 @@ class Solver {
 		si.upper_bound_p = Cp;
 		si.upper_bound_n = Cn;
 
-		System.out.print("\noptimization finished, #iter = "+iter+"\n");
+		svm.info("\noptimization finished, #iter = "+iter+"\n");
 	}
 
 	// return 1 if already optimal, return 0 otherwise
@@ -685,7 +707,7 @@ class Solver {
 					if (grad_diff > 0)
 					{
 						double obj_diff; 
-						double quad_coef=Q_i[i]+QD[j]-2*y[i]*Q_i[j];
+						double quad_coef=Q_i[i]+QD[j]-2.0*y[i]*Q_i[j];
 						if (quad_coef > 0)
 							obj_diff = -(grad_diff*grad_diff)/quad_coef;
 						else
@@ -709,7 +731,7 @@ class Solver {
 					if (grad_diff > 0)
 					{
 						double obj_diff; 
-						double quad_coef=Q_i[i]+QD[j]+2*y[i]*Q_i[j];
+						double quad_coef=Q_i[i]+QD[j]+2.0*y[i]*Q_i[j];
 						if (quad_coef > 0)
 							obj_diff = -(grad_diff*grad_diff)/quad_coef;
 						else
@@ -733,7 +755,7 @@ class Solver {
 		return 0;
 	}
 
-	private boolean be_shrunken(int i, double Gmax1, double Gmax2)
+	private boolean be_shrunk(int i, double Gmax1, double Gmax2)
 	{	
 		if(is_upper_bound(i))
 		{
@@ -790,15 +812,20 @@ class Solver {
 			}
 		}
 
-		// shrink
+		if(unshrink == false && Gmax1 + Gmax2 <= eps*10) 
+		{
+			unshrink = true;
+			reconstruct_gradient();
+			active_size = l;
+		}
 
 		for(i=0;i<active_size;i++)
-			if (be_shrunken(i, Gmax1, Gmax2))
+			if (be_shrunk(i, Gmax1, Gmax2))
 			{
 				active_size--;
 				while (active_size > i)
 				{
-					if (!be_shrunken(active_size, Gmax1, Gmax2))
+					if (!be_shrunk(active_size, Gmax1, Gmax2))
 					{
 						swap_index(i,active_size);
 						break;
@@ -806,28 +833,6 @@ class Solver {
 					active_size--;
 				}
 			}
-
-		// unshrink, check all variables again before final iterations
-
-		if(unshrinked || Gmax1 + Gmax2 > eps*10) return;
-
-		unshrinked = true;
-		reconstruct_gradient();
-
-		for(i=l-1;i>=active_size;i--)
-			if (!be_shrunken(i, Gmax1, Gmax2))
-			{
-				while (active_size < i)
-				{
-					if (be_shrunken(active_size, Gmax1, Gmax2))
-					{
-						swap_index(i,active_size);
-						break;
-					}
-					active_size++;
-				}
-				active_size++;
-			}
 	}
 
 	double calculate_rho()
@@ -989,7 +994,7 @@ final class Solver_NU extends Solver
 		}
 
 		if(Math.max(Gmaxp+Gmaxp2,Gmaxn+Gmaxn2) < eps)
- 			return 1;
+			return 1;
 	
 		if(y[Gmin_idx] == +1)
 			working_set[0] = Gmaxp_idx;
@@ -1000,7 +1005,7 @@ final class Solver_NU extends Solver
 		return 0;
 	}
 
-	private boolean be_shrunken(int i, double Gmax1, double Gmax2, double Gmax3, double Gmax4)
+	private boolean be_shrunk(int i, double Gmax1, double Gmax2, double Gmax3, double Gmax4)
 	{
 		if(is_upper_bound(i))
 		{
@@ -1049,15 +1054,20 @@ final class Solver_NU extends Solver
 			}
 		}
 
-		// shrinking
+		if(unshrink == false && Math.max(Gmax1+Gmax2,Gmax3+Gmax4) <= eps*10) 
+		{
+			unshrink = true;
+			reconstruct_gradient();
+			active_size = l;
+		}
 
 		for(i=0;i<active_size;i++)
-			if (be_shrunken(i, Gmax1, Gmax2, Gmax3, Gmax4))
+			if (be_shrunk(i, Gmax1, Gmax2, Gmax3, Gmax4))
 			{
 				active_size--;
 				while (active_size > i)
 				{
-					if (!be_shrunken(active_size, Gmax1, Gmax2, Gmax3, Gmax4))
+					if (!be_shrunk(active_size, Gmax1, Gmax2, Gmax3, Gmax4))
 					{
 						swap_index(i,active_size);
 						break;
@@ -1065,26 +1075,6 @@ final class Solver_NU extends Solver
 					active_size--;
 				}
 			}
-
-		if(unshrinked || Math.max(Gmax1+Gmax2,Gmax3+Gmax4) > eps*10) return;
-	
-		unshrinked = true;
-		reconstruct_gradient();
-
-		for(i=l-1;i>=active_size;i--)
-			if (!be_shrunken(i, Gmax1, Gmax2, Gmax3, Gmax4))
-			{
-				while (active_size < i)
-				{
-					if (be_shrunken(active_size, Gmax1, Gmax2, Gmax3, Gmax4))
-					{
-						swap_index(i,active_size);
-						break;
-					}
-					active_size++;
-				}
-				active_size++;
-			}
 	}
 	
 	double calculate_rho()
@@ -1160,10 +1150,10 @@ class SVC_Q extends Kernel
 	float[] get_Q(int i, int len)
 	{
 		float[][] data = new float[1][];
-		int start;
+		int start, j;
 		if((start = cache.get_data(i,data,len)) < len)
 		{
-			for(int j=start;j<len;j++)
+			for(j=start;j<len;j++)
 				data[0][j] = (float)(y[i]*y[j]*kernel_function(i,j));
 		}
 		return data[0];
@@ -1200,10 +1190,10 @@ class ONE_CLASS_Q extends Kernel
 	float[] get_Q(int i, int len)
 	{
 		float[][] data = new float[1][];
-		int start;
+		int start, j;
 		if((start = cache.get_data(i,data,len)) < len)
 		{
-			for(int j=start;j<len;j++)
+			for(j=start;j<len;j++)
 				data[0][j] = (float)kernel_function(i,j);
 		}
 		return data[0];
@@ -1263,10 +1253,10 @@ class SVR_Q extends Kernel
 	float[] get_Q(int i, int len)
 	{
 		float[][] data = new float[1][];
-		int real_i = index[i];
+		int j, real_i = index[i];
 		if(cache.get_data(real_i,data,l) < l)
 		{
-			for(int j=0;j<l;j++)
+			for(j=0;j<l;j++)
 				data[0][j] = (float)kernel_function(real_i,j);
 		}
 
@@ -1274,8 +1264,8 @@ class SVR_Q extends Kernel
 		float buf[] = buffer[next_buffer];
 		next_buffer = 1 - next_buffer;
 		byte si = sign[i];
-		for(int j=0;j<len;j++)
-			buf[j] = si * sign[j] * data[0][index[j]];
+		for(j=0;j<len;j++)
+			buf[j] = (float) si * sign[j] * data[0][index[j]];
 		return buf;
 	}
 
@@ -1289,6 +1279,21 @@ public class svm {
 	//
 	// construct and solve various formulations
 	//
+	public static final int LIBSVM_VERSION=289; 
+
+	public static svm_print_interface svm_print_string = new svm_print_interface()
+	{
+		public void print(String s)
+		{
+			System.out.print(s);
+		}
+	};
+
+	static void info(String s) 
+	{
+		svm_print_string.print(s);
+	}
+
 	private static void solve_c_svc(svm_problem prob, svm_parameter param,
 					double[] alpha, Solver.SolutionInfo si,
 					double Cp, double Cn)
@@ -1315,14 +1320,14 @@ public class svm {
 			sum_alpha += alpha[i];
 
 		if (Cp==Cn)
-			System.out.print("nu = "+sum_alpha/(Cp*prob.l)+"\n");
+			svm.info("nu = "+sum_alpha/(Cp*prob.l)+"\n");
 
 		for(i=0;i<l;i++)
 			alpha[i] *= y[i];
 	}
 
 	private static void solve_nu_svc(svm_problem prob, svm_parameter param,
-				 	double[] alpha, Solver.SolutionInfo si)
+					double[] alpha, Solver.SolutionInfo si)
 	{
 		int i;
 		int l = prob.l;
@@ -1361,7 +1366,7 @@ public class svm {
 			alpha, 1.0, 1.0, param.eps, si, param.shrinking);
 		double r = si.r;
 
-		System.out.print("C = "+1/r+"\n");
+		svm.info("C = "+1/r+"\n");
 
 		for(i=0;i<l;i++)
 			alpha[i] *= y[i]/r;
@@ -1373,7 +1378,7 @@ public class svm {
 	}
 
 	private static void solve_one_class(svm_problem prob, svm_parameter param,
-				    	double[] alpha, Solver.SolutionInfo si)
+					double[] alpha, Solver.SolutionInfo si)
 	{
 		int l = prob.l;
 		double[] zeros = new double[l];
@@ -1430,7 +1435,7 @@ public class svm {
 			alpha[i] = alpha2[i] - alpha2[i+l];
 			sum_alpha += Math.abs(alpha[i]);
 		}
-		System.out.print("nu = "+sum_alpha/(param.C*l)+"\n");
+		svm.info("nu = "+sum_alpha/(param.C*l)+"\n");
 	}
 
 	private static void solve_nu_svr(svm_problem prob, svm_parameter param,
@@ -1460,7 +1465,7 @@ public class svm {
 		s.Solve(2*l, new SVR_Q(prob,param), linear_term, y,
 			alpha2, C, C, param.eps, si, param.shrinking);
 
-		System.out.print("epsilon = "+(-si.r)+"\n");
+		svm.info("epsilon = "+(-si.r)+"\n");
 		
 		for(i=0;i<l;i++)
 			alpha[i] = alpha2[i] - alpha2[i+l];
@@ -1500,7 +1505,7 @@ public class svm {
 				break;
 		}
 
-		System.out.print("obj = "+si.obj+", rho = "+si.rho+"\n");
+		svm.info("obj = "+si.obj+", rho = "+si.rho+"\n");
 
 		// output SVs
 
@@ -1524,7 +1529,7 @@ public class svm {
 			}
 		}
 
-		System.out.print("nSV = "+nSV+", nBSV = "+nBSV+"\n");
+		svm.info("nSV = "+nSV+", nBSV = "+nBSV+"\n");
 
 		decision_function f = new decision_function();
 		f.alpha = alpha;
@@ -1544,7 +1549,7 @@ public class svm {
 			if (labels[i] > 0) prior1+=1;
 			else prior0+=1;
 	
-		int max_iter=100; 	// Maximal number of iterations
+		int max_iter=100;	// Maximal number of iterations
 		double min_step=1e-10;	// Minimal step taken in line search
 		double sigma=1e-12;	// For numerically strict PD of Hessian
 		double eps=1e-5;
@@ -1608,7 +1613,7 @@ public class svm {
 			gd=g1*dA+g2*dB;
 
 
-			stepsize = 1; 		// Line Search
+			stepsize = 1;		// Line Search
 			while (stepsize >= min_step)
 			{
 				newA = A + stepsize * dA;
@@ -1636,13 +1641,13 @@ public class svm {
 			
 			if (stepsize < min_step)
 			{
-				System.err.print("Line search fails in two-class probability estimates\n");
+				svm.info("Line search fails in two-class probability estimates\n");
 				break;
 			}
 		}
 		
 		if (iter>=max_iter)
-			System.err.print("Reaching maximal iterations in two-class probability estimates\n");
+			svm.info("Reaching maximal iterations in two-class probability estimates\n");
 		probAB[0]=A;probAB[1]=B;
 	}
 
@@ -1712,7 +1717,7 @@ public class svm {
 			}
 		}
 		if (iter>=max_iter)
-			System.err.print("Exceeds max_iter in multiclass_prob\n");
+			svm.info("Exceeds max_iter in multiclass_prob\n");
 	}
 
 	// Cross-validation decision values for probability estimates
@@ -1822,7 +1827,7 @@ public class svm {
 			else 
 				mae+=Math.abs(ymv[i]);
 		mae /= (prob.l-count);
-		System.err.print("Prob. model for test data: target value = predicted value + z,\nz: Laplace distribution e^(-|z|/sigma)/(2sigma),sigma="+mae+"\n");
+		svm.info("Prob. model for test data: target value = predicted value + z,\nz: Laplace distribution e^(-|z|/sigma)/(2sigma),sigma="+mae+"\n");
 		return mae;
 	}
 
@@ -2071,7 +2076,7 @@ public class svm {
 				nz_count[i] = nSV;
 			}
 
-			System.out.print("Total nSV = "+nnz+"\n");
+			svm.info("Total nSV = "+nnz+"\n");
 
 			model.l = nnz;
 			model.SV = new svm_node[nnz][];
@@ -2139,7 +2144,7 @@ public class svm {
 			int nr_class = tmp_nr_class[0];
 			int[] label = tmp_label[0];
 			int[] start = tmp_start[0];
-			int[] count = tmp_count[0]; 		
+			int[] count = tmp_count[0];		
 
 			// random shuffle and then data grouped by fold using the array perm
 			int[] fold_count = new int[nr_fold];
@@ -2396,7 +2401,7 @@ public class svm {
 
 	public static void svm_save_model(String model_file_name, svm_model model) throws IOException
 	{
-		DataOutputStream fp = new DataOutputStream(new FileOutputStream(model_file_name));
+		DataOutputStream fp = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(model_file_name)));
 
 		svm_parameter param = model.param;
 
