@@ -2,9 +2,12 @@
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
+#include <errno.h>
 #include "svm.h"
 #define Malloc(type,n) (type *)malloc((n)*sizeof(type))
 
+void print_null(const char *s) {}
+
 void exit_with_help()
 {
 	printf(
@@ -30,14 +33,21 @@ void exit_with_help()
 	"-p epsilon : set the epsilon in loss function of epsilon-SVR (default 0.1)\n"
 	"-m cachesize : set cache memory size in MB (default 100)\n"
 	"-e epsilon : set tolerance of termination criterion (default 0.001)\n"
-	"-h shrinking: whether to use the shrinking heuristics, 0 or 1 (default 1)\n"
-	"-b probability_estimates: whether to train a SVC or SVR model for probability estimates, 0 or 1 (default 0)\n"
-	"-wi weight: set the parameter C of class i to weight*C, for C-SVC (default 1)\n"
+	"-h shrinking : whether to use the shrinking heuristics, 0 or 1 (default 1)\n"
+	"-b probability_estimates : whether to train a SVC or SVR model for probability estimates, 0 or 1 (default 0)\n"
+	"-wi weight : set the parameter C of class i to weight*C, for C-SVC (default 1)\n"
 	"-v n: n-fold cross validation mode\n"
+	"-q : quiet mode (no outputs)\n"
 	);
 	exit(1);
 }
 
+void exit_input_error(int line_num)
+{
+	fprintf(stderr,"Wrong input format at line %d\n", line_num);
+	exit(1);
+}
+
 void parse_command_line(int argc, char **argv, char *input_file_name, char *model_file_name);
 void read_problem(const char *filename);
 void do_cross_validation();
@@ -49,6 +59,27 @@ struct svm_node *x_space;
 int cross_validation;
 int nr_fold;
 
+static char *line = NULL;
+static int max_line_len;
+
+static char* readline(FILE *input)
+{
+	int len;
+	
+	if(fgets(line,max_line_len,input) == NULL)
+		return NULL;
+
+	while(strrchr(line,'\n') == NULL)
+	{
+		max_line_len *= 2;
+		line = (char *) realloc(line,max_line_len);
+		len = (int) strlen(line);
+		if(fgets(line+len,max_line_len-len,input) == NULL)
+			break;
+	}
+	return line;
+}
+
 int main(int argc, char **argv)
 {
 	char input_file_name[1024];
@@ -79,6 +110,7 @@ int main(int argc, char **argv)
 	free(prob.y);
 	free(prob.x);
 	free(x_space);
+	free(line);
 
 	return 0;
 }
@@ -188,6 +220,10 @@ void parse_command_line(int argc, char **argv, char *input_file_name, char *mode
 			case 'b':
 				param.probability = atoi(argv[i]);
 				break;
+			case 'q':
+				svm_print_string = &print_null;
+				i--;
+				break;
 			case 'v':
 				cross_validation = 1;
 				nr_fold = atoi(argv[i]);
@@ -205,7 +241,7 @@ void parse_command_line(int argc, char **argv, char *input_file_name, char *mode
 				param.weight[param.nr_weight-1] = atof(argv[i]);
 				break;
 			default:
-				fprintf(stderr,"unknown option\n");
+				fprintf(stderr,"Unknown option: -%c\n", argv[i-1][1]);
 				exit_with_help();
 		}
 	}
@@ -234,9 +270,11 @@ void parse_command_line(int argc, char **argv, char *input_file_name, char *mode
 
 void read_problem(const char *filename)
 {
-	int elements, max_index, i, j;
+	int elements, max_index, inst_max_index, i, j;
 	FILE *fp = fopen(filename,"r");
-	
+	char *endptr;
+	char *idx, *val, *label;
+
 	if(fp == NULL)
 	{
 		fprintf(stderr,"can't open input file %s\n",filename);
@@ -245,25 +283,24 @@ void read_problem(const char *filename)
 
 	prob.l = 0;
 	elements = 0;
-	while(1)
+
+	max_line_len = 1024;
+	line = Malloc(char,max_line_len);
+	while(readline(fp)!=NULL)
 	{
-		int c = fgetc(fp);
-		switch(c)
+		char *p = strtok(line," \t"); // label
+
+		// features
+		while(1)
 		{
-			case '\n':
-				++prob.l;
-				// fall through,
-				// count the '-1' element
-			case ':':
-				++elements;
+			p = strtok(NULL," \t");
+			if(p == NULL || *p == '\n') // check '\n' as ' ' may be after the last feature
 				break;
-			case EOF:
-				goto out;
-			default:
-				;
+			++elements;
 		}
+		++elements;
+		++prob.l;
 	}
-out:
 	rewind(fp);
 
 	prob.y = Malloc(double,prob.l);
@@ -274,33 +311,43 @@ out:
 	j=0;
 	for(i=0;i<prob.l;i++)
 	{
-		double label;
+		inst_max_index = -1; // strtol gives 0 if wrong format, and precomputed kernel has <index> start from 0
+		readline(fp);
 		prob.x[i] = &x_space[j];
-		fscanf(fp,"%lf",&label);
-		prob.y[i] = label;
+		label = strtok(line," \t");
+		prob.y[i] = strtod(label,&endptr);
+		if(endptr == label)
+			exit_input_error(i+1);
 
 		while(1)
 		{
-			int c;
-			do {
-				c = getc(fp);
-				if(c=='\n') goto out2;
-			} while(isspace(c));
-			ungetc(c,fp);
-			if (fscanf(fp,"%d:%lf",&(x_space[j].index),&(x_space[j].value)) < 2)
-			{
-				fprintf(stderr,"Wrong input format at line %d\n", i+1);
-				exit(1);
-			}
+			idx = strtok(NULL,":");
+			val = strtok(NULL," \t");
+
+			if(val == NULL)
+				break;
+
+			errno = 0;
+			x_space[j].index = (int) strtol(idx,&endptr,10);
+			if(endptr == idx || errno != 0 || *endptr != '\0' || x_space[j].index <= inst_max_index)
+				exit_input_error(i+1);
+			else
+				inst_max_index = x_space[j].index;
+
+			errno = 0;
+			x_space[j].value = strtod(val,&endptr);
+			if(endptr == val || errno != 0 || (*endptr != '\0' && !isspace(*endptr)))
+				exit_input_error(i+1);
+
 			++j;
-		}	
-out2:
-		if(j>=1 && x_space[j-1].index > max_index)
-			max_index = x_space[j-1].index;
+		}
+
+		if(inst_max_index > max_index)
+			max_index = inst_max_index;
 		x_space[j++].index = -1;
 	}
 
-	if(param.gamma == 0)
+	if(param.gamma == 0 && max_index > 0)
 		param.gamma = 1.0/max_index;
 
 	if(param.kernel_type == PRECOMPUTED)
